// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/inotify.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <unistd.h>
#include <netinet/in.h>
#include "server.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <openssl/md5.h>

#include "lock.h"

#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

static int stalltime=100;
static int server_jobs_run=0;
static double server_jobs_per_s=0.0;
static int server_odes=0;
static double server_odes_per_s=0.0;

static int stalled=FALSE;

void alarm_wakeup (int i)
{
	struct itimerval tout_val;

	signal(SIGALRM,alarm_wakeup);

	tout_val.it_interval.tv_sec = 0;
	tout_val.it_interval.tv_usec = 0;
	tout_val.it_value.tv_sec = stalltime; /* 10 seconds timer */
	tout_val.it_value.tv_usec = 0;

	if (stalled==TRUE)
	{
		printf("\nJob stalled.\n");
		exit(0);
	}else
	{
		stalled=TRUE;
		server_jobs_per_s=((double)server_jobs_run)/((double)stalltime);
		server_jobs_run=0;
		server_odes_per_s=((double)server_odes)/((double)stalltime);
		server_odes=0;

	}

	setitimer(ITIMER_REAL, &tout_val,0);

}


int cmp_lock(char *in)
{
char temp[1000];
int max=strlen("lock");
if (strlen(in)>max)
{
strcpy(temp,in);
temp[max]=0;
return strcmp("lock",temp);
}
return -1;
}
void server_shut_down(struct server *myserver)
{
if (myserver->cpus!=0)
{
	inotify_rm_watch( myserver->fd, myserver->wd );
	close( myserver->fd );
}
printf("I have shut down the server\n");
myserver->cpus=0;
myserver->on=FALSE;
}

void print_jobs(struct server *myserver)
{
int i;

for (i=0;i<server_max;i++)
{
	if (myserver->state[i]!=0)
	{
		printf("%d %s %d\n",i,myserver->command[i],myserver->state[i]);
	}
}
}

double server_get_odes_per_s()
{
return server_odes_per_s;
}

double server_get_jobs_per_s()
{
return server_jobs_per_s;
}

void server_init(struct server *myserver)
{
int i;
char name[200];
int hpc_cpus;
myserver->on=TRUE;

struct itimerval tout_val;



if (myserver->readconfig==TRUE)
{
	FILE *config=fopen("./server.inp","r");
	if (config==NULL)
	{
		printf("Can not find server.inp\n");
		exit(0);
	}
	fscanf(config,"%s",name);
	fscanf(config,"%d",&(myserver->cpus));
	fscanf(config,"%s",name);
	fscanf(config,"%d",&(hpc_cpus));

	#ifdef hpc
	myserver->cpus=hpc_cpus;
	#endif

	fscanf(config,"%s",name);
	fscanf(config,"%d",&(stalltime));

	tout_val.it_interval.tv_sec = 0;
	tout_val.it_interval.tv_usec = 0;
	tout_val.it_value.tv_sec = stalltime; /* 10 seconds timer */
	tout_val.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &tout_val,0);

	signal(SIGALRM,alarm_wakeup); /* set the Alarm signal capture */


	fclose(config);
}

myserver->jobs=0;
myserver->jobs_running=0;

if (myserver->cpus>0)
{
	for (i=0;i<server_max;i++)
	{
		myserver->state[i]=server_no_job;
		strcpy(myserver->command[i],"");
	}


	myserver->fd=inotify_init();

	if ( myserver->fd < 0 )
	{
		perror( "inotify_init" );
	}

	myserver->wd = inotify_add_watch( myserver->fd, "./", IN_CREATE | IN_CLOSE_WRITE );
}

}


int server_decode(char *command)
{
int odes=0;
	if (cmpstr_min(command,"gendosn")==0)
	{
		gen_dos_fd_gaus_n(extract_str_number(command,"gendosn"));
	}else
	if (cmpstr_min(command,"gendosp")==0)
	{
		gen_dos_fd_gaus_p(extract_str_number(command,"gendosp"));
	}else
	{
		odes=run_simulation(command);
	}
return odes;
}

void server_add_job(struct server *myserver,char *command)
{

if (myserver->on==FALSE)
{
	printf("No server running!\n");
	exit(0);
}
if (myserver->cpus>0)
{
int i;
	for (i=0;i<server_max;i++)
	{
		if (myserver->state[i]==server_no_job)
		{
			myserver->state[i]=server_job_ready;
			strcpy(myserver->command[i],command);
			myserver->jobs++;
			break;
		}
	}
}else
{
	server_decode(command);
}

}

void server_exe_jobs(struct server *myserver)
{
if (myserver->jobs==0) return;
if (myserver->jobs_running>=myserver->cpus) return;
int i;

	for (i=0;i<server_max;i++)
	{
		if (myserver->state[i]==server_job_ready)
		{

			if (myserver->jobs_running<myserver->cpus)
			{
				myserver->state[i]=server_job_running;
				printf("just run %s\n",myserver->command[i]);
				myserver->jobs_running++;
				//printf("I would have run\n");
				//printf("%d %d\n",myserver->jobs_running,myserver->cpus);
				//getchar();

				signal(SIGCHLD, SIG_IGN);
				pid_t pID = fork();
				if (pID == 0)
				{
					int ode=0;
					ode=server_decode(myserver->command[i]);
					//int i=0;
					//do
					//{
					//i++;
					//}while(i<1e6);


					char lockname[500];
					sprintf(lockname,"./lock%d.dat",i);
					FILE *out=fopen(lockname,"w");
					fprintf(out,"%s\n%d\n",myserver->command[i],ode);
					fclose(out);

					exit(0);
				}
			}else
			{
				break;
			}



		}
	}
}


void server_job_finished(struct server *myserver,char *job)
{

int i;
for (i=0;i<server_max;i++)
{
	if (strcmp(myserver->command[i],job)==0)
	{
		if (myserver->state[i]!=server_no_job)
		{
			myserver->state[i]=server_no_job;
			strcpy(myserver->command[i],"");
			myserver->jobs_running--;
			myserver->jobs--;
		}
		break;
	}
}

}
void server_run_jobs(struct server *myserver)
{

int length;
int i=0;
char buffer[EVENT_BUF_LEN];
if (myserver->jobs==0) return;

if (myserver->cpus>0)
{
	printf("server_run_jobs\n");
	server_exe_jobs(myserver);

do
{
	length = read( myserver->fd, buffer, EVENT_BUF_LEN );

	if ( length < 0 )
	{
		perror( "read" );
	}

	while ( i < length )
	{

		struct inotify_event *event = ( struct inotify_event * ) &buffer[ i ];
		if ( event->len )
		{
			if ( event->mask & IN_CLOSE_WRITE )
			{
					if (cmp_lock(event->name)==0)
					{
						char buf[200];
						//FILE *mes=fopen("./mes.dat","a");
						//fprintf(mes,"\nrecived message: %s\n\n", buf);
						//fclose(mes);
						//getchar();
						FILE *lock=fopen(event->name,"r");
						int ode;
						fscanf(lock,"%s",buf);
						fscanf(lock,"%d",&ode);
						fclose(lock);

						server_odes+=ode;
						server_jobs_run+=1;
						server_job_finished(myserver,buf);
						server_exe_jobs(myserver);

						stalled=FALSE;

						print_jobs(myserver);
						if (myserver->jobs==0)
						{
						printf("All jobs finished\n");
						return;
						}

						if (myserver->jobs_running==0)
						{
							printf("I have finished all my jobs\n");
							break;
						}
					}
			}
		}
		i += EVENT_SIZE + event->len;



	}
	i=0;




}while(1);

}
}

